local segger_rtt_port = 19023;
local function bkpt_file_id(filename)
	local new_path = edx:regular_path(filename);
	if (new_path) then
		return new_path;
	end;
	return filename:gsub("^%w:", function(x)
			return x:upper();
	end);
end;

-- load all breakpoints in the project_configs and opened documents
local function load_breakpoints(self)
	local all_doc = {mgr.get_all_documents()};
	local store_breakpoints = project_configs and (project_configs[".breakpoints"] or {});
	local breakpoints = {};
	if project:is_ready() then
		for short_name, info in pairs(store_breakpoints) do
			local file_bkps = {};
			local file_id = bkpt_file_id(project:make_path(short_name));
			breakpoints[file_id] = file_bkps;
			for line, info in pairs(info) do
				table.insert(file_bkps, line);
			end;
		end;
	end;

	for idx, doc in pairs(all_doc) do
		if doc.doc_type ~= doc_type.hex then
			local bkps = {doc.get_all_breakpoints()};
			local name = bkpt_file_id(doc.file_name);
			if #bkps <=0 then
				breakpoints[name] = nil;
			else
				local file_bkps = {};
				breakpoints[name] = file_bkps;
				for idx, line in pairs(bkps) do
					table.insert(file_bkps, line);
				end;
			end;
		end;
	end;
	self.bkps = {};
	for filename, bpk_info in pairs(breakpoints) do
		for idx, line in pairs(bpk_info) do
			if self.bkps[filename] == nil then
				self.bkps[filename] = {};
			end;
			line = math.tointeger(line);
			local bkp_id = self:set_breakpoint_mapped(filename, line + 1);
			if bkp_id == 0 then
				-- doc:toggle_breakpoint( line );
				print("Breakpoint disabled:", filename, ":", line);
			else
				print("Breakpoint load:", filename, ":", line+1);
				self.bkps[filename][tostring(line+1)] = bkp_id;
			end;
		end;
	end;
end;

local function remove_breakpoints(filename)
	filename = bkpt_file_id(make_path(filename));
	local doc = mgr:get_document(filename);
	if doc ~= nil then
		-- remove breakpoints in the opened document
		local bkps = {doc.get_all_breakpoints()};
		for i, ln in ipairs(bkps) do
			doc:toggle_breakpoint(ln);
		end;
	end;

	local store_breakpoints = project_configs and project_configs[".breakpoints"];
	if store_breakpoints ~= nil and project:is_ready() then
		-- remove breakpoints in project_configs
		local project_home = project:home();
		if is_file_in_path(filename, project_home) then
			local short_name = filename:sub(2 + #project_home);
			short_name = short_name:gsub('\\', '/');
			store_breakpoints[short_name] = nil;
		end;
	end;

	local dbg = edx.dbg;
	if dbg ~= nil and dbg.bkps[filename] ~= nil then
		local bkps = dbg.bkps[filename];
		dbg.bkps[filename] = nil;
		for ln, id in pairs(bkps) do
			dbg:del_breakpoint(id);
		end;
	end;
end;

function list_breakpoints()
	local all_doc = {mgr.get_all_documents()};
	local store_breakpoints = project_configs and (project_configs[".breakpoints"] or {});
	local breakpoints = {};
	if project:is_ready() then
		for short_name, info in pairs(store_breakpoints) do
			local file_bkps = {};
			local file_id = bkpt_file_id(project:make_path(short_name));
			breakpoints[file_id] = file_bkps;
			for line, info in pairs(info) do
				table.insert(file_bkps, line);
			end;
		end;
	end;

	for idx, doc in pairs( all_doc ) do
		local bkps = {doc.get_all_breakpoints()};
		local name = bkpt_file_id(doc.file_name);
		if #bkps <=0 then
			breakpoints[name] = nil;
		else
			local file_bkps = {};
			breakpoints[name] = file_bkps;
			for idx, line in pairs(bkps) do
				table.insert(file_bkps, math.floor(line));
			end;
		end;
	end;

	edx:clear("output");
	print("breakpoints:");
	for name, info in pairs(breakpoints) do
		for i, ln in ipairs(info) do
			print(name, ":", (ln+1), ":");
		end;
	end;
	edx:active("output");
end;

local function dbg_on_exception(self, code, info)
	print(("on exception 0x%08X - %s"):format(code, info));
	if info == "BREAKPOINT" or info:find("SIGTRAP") then
		show_notify(LANG("notify/process_suspended"));
	else
		show_notify(("Exception 0x%08X - %s"):format(code, info));
	end;
	local frames = {self:get_callstack()};
	for idx, frame in ipairs(frames) do
		local src = self.unmap_path(frame.file);
		local line = frame.line;
		if src and utils:file_time(src) and self:select_frame(frame.id) then
			goto success;
		end;
	end;
	self:locate_file();
	::success::
	edx:active( "" );
	menu_bar:switch_debug_menu("paused");
	return 1;
end;

local function dbg_on_step(self)
	self.__trace_count = 0;
	-- print("step");
	self:locate_file();
	edx:active("");
	menu_bar:switch_debug_menu("paused");
	return 1;
end;

local function dbg_on_trace(self)
	local fp = {self:get_src_line()};
	local src = fp[1];
	local line = fp[2];
	if self.__trace_count == nil then
		self.__trace_count = 1;
	else
		self.__trace_count = self.__trace_count + 1;
	end;
	if self.__trace_count >= 10 then
		print( "on trace - too much times" );
	elseif src == nil then
		print( "on trace - again" );
		self:trace();
		return 1;
	elseif self.last_fp[2] == tonumber(line) and self.last_fp[3] == src then
		print( "on trace - again2" );
		self:trace();
		return 1;
	end;
	-- print( "on trace" );
	self:locate_file();
	edx:active("");
	menu_bar:switch_debug_menu("paused");
	return 1;
end;

local function dbg_on_trace_return(self)
	self.__trace_count = 0;
	-- print( "on trace return" );
	self:locate_file();
	edx:active("");
	menu_bar:switch_debug_menu("paused");
	return 1;
end;

local function dbg_on_breakpoint(self, id, file, lineno)
	self.__trace_count = 0;
	file = self.unmap_path(file);
	file = bkpt_file_id(file);
	print(("on breakpoint(%d) %s:%d"):format(id, file, lineno));
	local bkps = self.bkps[file] or {};
	local bkps_id = bkps[tostring(lineno)];
	if bkps_id ~= id then
		print("breakpoint not exists, remove the unknown breakpoint!");
		self:del_breakpoint(id);
	else
		local doc = mgr:get_document(file);
		if doc ~= nil and not doc:has_breakpoint(lineno-1) then
			print("restore missing breakpoint!");
			doc:toggle_breakpoint( lineno-1 );
		end;
	end;

	self:locate_file();
	edx:active("");
	menu_bar:switch_debug_menu("paused");
	return 1;
end;

local function dbg_on_load(self)
	self.__trace_count = 0;
	-- edx:clear("output");
	edx:switch_layout("debug");
	menu_bar:switch_debug_menu( "debugging" );
	print("debug session start");
	if self._post_load_handler and #self._post_load_handler > 0 then
		for idx, handler in ipairs(self._post_load_handler) do
			pinvoke(handler, self);
		end;
	end;
	self:load_breakpoints();
	if self._on_load_go then
		self:_on_load_go();
		self._on_load_go = nil;
	else
		self:go();
	end;
	edx:update_dbg_info();
	return 1;
end;

local function dbg_on_terminate(self, code)
	if self.last_fp ~= nil then
		self.last_fp[1].set_dbg_cursor(-1);
	end;
	edx:clear_dbg_info();
	edx.dbg = nil;
	edx:active( "" );
	print( ("process exit(0x%08X)"):format(code) );
	print( "debug session end" );
	if xws_mgr.is_loaded then
		edx:switch_layout("simple");
	else
		edx:switch_layout("lite");
	end;
	menu_bar:switch_debug_menu("stopped");
	
	local peripherals = edx:get_object("pane:peripherals");
	peripherals:unload_svd();
	-- release debugger object
	edx:select_debugger("");
	return 1;
end;

local function dbg_on_pause(self)
	self.__trace_count = 0;
	-- print("on pause");
	self:locate_file();
	menu_bar:switch_debug_menu("paused");
	return 1;
end;

local function dbg_on_stack_update(self)
	-- print("on stack update");
	self:locate_file();
	return 1;
end;

local function dbg_on_thread_switched(self)
	-- print("on thread switched");
	self:locate_file();
	return 1;
end;

local function dbg_on_message(self, type_id, msg)
	if type_id == 1 then
		return 0;	-- console message
	end;
	if type_id == 2 and (
			msg:match("^%s*rtt:%s+Control%s+block%s+found.*$") or 
			msg:match("^%s*Found%s+RTT%s+control%s+block.*$")
		) then
		-- segger rtt found!
		self._on_locate_file = nil;	-- remove segger rtt handler
		self:exec(("mon rtt server start %u 0"):format(segger_rtt_port));	-- start segger rtt server
		self:set_param("ENABLE_SEGGER_RTT_MONITOR", tostring(segger_rtt_port));
	end;
	if type_id == 4 then
		-- segger-rtt message
		edx:output(msg);
		return 1;
	end;
	if type_id == 3 then
		edx:output(msg);
	end;
	local type_prefix = {
		[2] = "\x1b[32;44m@\x1b[33;44m ";	-- remote response
		[3] = "\x1b[35;44m&\x1b[33;44m ";	-- debugger message
	};
	msg = msg:gsub("^(.-\n)", function(info)
			return type_prefix[type_id]..info.."\x1b[0m";
	end);
	self:print(msg);
	return 1;
end;

local function locate_file(self)
	if self._on_locate_file then
		pinvoke(self._on_locate_file, self);
	end;
	edx:update_dbg_info();
	local fp = self.fp;
	self.fp = nil;
	if fp == nil then
		fp = {self:get_src_line()};
	end;
	local src = self.unmap_path(fp[1]);
	local line = fp[2];
	if src == nil then
		print( "no source file" );
		return -2;
	end;
	if mgr:open_doc(src) then
		if self.last_fp ~= nil then
			self.last_fp[1].set_dbg_cursor(-1);
		end;
		local doc = mgr.current_document;
		doc:set_cursor(tonumber(line)-1, 0);
		self.last_fp = {doc, tonumber(line), src};
		doc:set_dbg_cursor(tonumber(line)-1);
		edx:do_cmd(_op.center_cursor_line);
		return 0;
	else
		print( ("can't open source file - %s:%s"):format( tostring(src), tostring(line) ) );
		return -1;
	end;
end;

function get_dbg()
	if edx.dbg == nil then
		local dbg_obj = edx.get_debugger();
		dbg_obj.load_breakpoints = load_breakpoints;
		dbg_obj.on_exception = dbg_on_exception;
		dbg_obj.on_step = dbg_on_step;
		dbg_obj.on_trace = dbg_on_trace;
		dbg_obj.on_trace_return = dbg_on_trace_return;
		dbg_obj.on_breakpoint = dbg_on_breakpoint;
		dbg_obj.on_load = dbg_on_load;
		dbg_obj.on_terminate = dbg_on_terminate;
		dbg_obj.on_pause = dbg_on_pause;
		dbg_obj.on_stack_update = dbg_on_stack_update;
		dbg_obj.on_thread_switched = dbg_on_thread_switched;
		dbg_obj.on_message = dbg_on_message;
		dbg_obj.locate_file = locate_file;
		dbg_obj.map_path = function(path) return path; end;
		dbg_obj.unmap_path = function(path) return path; end;
		dbg_obj.set_breakpoint_mapped = function(self, file, line)
			return self:set_breakpoint(self.map_path(file), line);
		end;
		dbg_obj.on_notify = function(self, event, info)
			-- print("on_notify:", event, " ", info);
			if event == "create thread" or event == "exit thread" then
				edx:update_dbg_info_ex("threads");
			elseif event == "load module" or event == "unload module" then
				edx:update_dbg_info_ex("modules");
			elseif event == "load error" then
				show_notify(LANG("notify/debug/can_not_debug_target") .. "\n" .. info);
				self:terminate();
			elseif event == "rtt" then
				if info == "connected" then
					edx:output("Segger RTT Server connected!\n");
				end;
			end;
		end;
		edx.dbg = dbg_obj;
		return dbg_obj;
	end;
	return edx.dbg;
end;

function clear_dbg()
	local dbg_obj = edx.dbg;
	if dbg_obj == nil then
		return;
	end;
	if dbg_obj.last_fp ~= nil then
		dbg_obj.last_fp[1].set_dbg_cursor(-1);
	end;
	edx.dbg = nil;
	print( "debug session end" );
end;

function load_program(post_load_handlers)
	project:reload_settings();
	local target_name = nil;
	local skip_current_doc = false;
	if edx:is_focused("solution") then
		local item_path, item_type, item_id = xws_mgr.get_selected_item();
		if item_type == 0x11 then
			-- is the target item
			target_name = xws_mgr:get_item_name(item_id);
		end;
	end;
	local cmd = project:find_build_command(target_name, skip_current_doc);
	local toolset = cmd and cmd.toolset;
	local dbg_target = _G["dbg_target"];
	local dbg_environment = nil;
	local dbg_working_path = nil;
	local dbg_argument = nil;
	local dbg_type = _G["dbg_type"] or "windbg";
	local embeded_dbg_servers = {["openocd"]=true, ["jlink"]=true, ["stlink"]=true, ["pyocd"]=true, ["t-head"]=true};
	if cmd and cmd.target_output then
		dbg_target = cmd.target_output;
		dbg_type = cmd.debug_type;
		if embeded_dbg_servers[dbg_type] then
			dbg_type = "gdb/mi";
		elseif dbg_type == "zephyr" then
			dbg_type = "gdb/mi";
		elseif dbg_type == "auto" then
			local symbol_type = edx:symbol_type(dbg_target);
			if symbol_type == "pdb" then
				dbg_type = "windbg";
				print("Found PDB info, debug with windbg~")
			else
				dbg_type = "gdb/mi";
			end;
		end;
		edx:select_debugger(dbg_type);
		edx.dbg = nil;
		dbg_argument = cmd.argument;
		dbg_working_path = cmd.cwd;
		dbg_environment = cmd.environment;
		if toolset and not toolset.wsl and not toolset.remote then
			-- add all target's path to PATH environment
			project:fix_path_env(dbg_environment);
		end;
	end;
	local map_path = (toolset and toolset.map_path) or function(path) return path; end;
	local unmap_path = (toolset and toolset.unmap_path) or function(path) return path; end;
	print("debug:", dbg_type, ":", dbg_target);
	local dbg = get_dbg();
	if post_load_handlers == nil then
		post_load_handlers = {};
	end;
	dbg._post_load_handler = post_load_handlers;

	dbg.map_path = map_path;
	dbg.unmap_path = unmap_path;
	dbg.dbg_type = dbg_type;
	if toolset and toolset["type"] == "llvm-mingw" and dbg_type == "windbg" then
		if libcxx_beautifier == nil then
			require "libcxx_beautifier";
		end;
		dbg.beautifier = edx.get_debugger_beautifier("libc++");
	else
		dbg.beautifier = edx.get_debugger_beautifier(dbg_type);
	end;
	if dbg_environment then
		for key, value in pairs(dbg_environment) do
			dbg:set_environment(key, value);
		end;
	end;
	if dbg_working_path then
		if dbg_working_path:sub(-1) == "\\" and #dbg_working_path > 3 then
			dbg_working_path = dbg_working_path:sub(1,-2);
		end;
		if toolset and toolset.wsl then
			dbg_working_path = map_path(dbg_working_path);
			dbg:set_working_path(dbg_working_path);
		elseif utils:file_time(dbg_working_path) or dbg_working_path:match([[^%w:\$]]) then
			dbg_working_path = map_path(dbg_working_path);
			dbg:set_working_path(dbg_working_path);
		else
			print("ERROR: debugee working directory not exists! ", dbg_working_path);
			return false;
		end;
	end;
	if dbg_argument then
		if type(dbg_argument) == "table" then
			dbg_argument = table.concat(dbg_argument, " ");
		end;
		dbg:set_args(dbg_argument);
	end;
	if dbg_type:match("gdb/") then
		local gdb_post_load_handler = function()
			if cmd then
				print("debug target:", cmd.target_output, " ", cmd.argument);
				if cmd.cwd then
					print("debug cwd:", cmd.cwd);
				end;
			end;
			if toolset then
				if toolset.encoding then
					dbg:set_param("LOCAL_CHARSET", toolset.encoding);
					print("debug encoding:", toolset.encoding);
				else
					dbg:set_param("LOCAL_CHARSET", "");
				end;
				if toolset.wsl then
					dbg:set_param("ENABLE_WSL_CONSOLE", toolset.wsl);
				elseif toolset.remote then
					dbg:set_param("ENABLE_SSH_CONSOLE", toolset.remote);
				end;
			end;
			print("load gdb-pretty-print...");
			local pretty_print_path;
			if toolset and toolset.tools and (toolset.tools.gcc or toolset.tools.clang or toolset["type"] == "esp-idf") then
				if toolset.get_pretty_print_path then
					dbg_call(function()
							pretty_print_path = toolset:get_pretty_print_path();
					end, "");
				else
					pretty_print_path = toolset.tools["gcc-pretty-print"] or toolset.tools["libcxx-pretty-print"];
				end;
			end;
			if pretty_print_path then
				pretty_print_path = pretty_print_path:gsub("\\", "/");
				print("gdb-pretty-print:", pretty_print_path);
				if toolset.tools.gcc then
					-- gcc mode
					dbg:exec(
						"python\n"..
						"import sys\n"..
						"sys.path.insert(0, '"..pretty_print_path.."')\n"..
						"from libstdcxx.v6.printers import register_libstdcxx_printers\n"..
						"register_libstdcxx_printers(None)\n"..
						"end\n"
					);
				elseif toolset.tools.clang then
					-- clang mode
					dbg:exec(
						"source "..pretty_print_path.."\n"..
						"python register_libcxx_printer_loader()\n"
					);
				end;
				dbg:exec("-enable-pretty-printing\n");
			else
				print("gdb-pretty-print disabled!");
			end;
		end;
		table.insert(dbg._post_load_handler, gdb_post_load_handler);
		if toolset and toolset.tools and toolset.tools.gdb then
			dbg:set_param("GDB_PATH", toolset.tools.gdb);
		elseif GDB_PATHS and GDB_PATHS[1] then
			dbg:set_param("GDB_PATH", GDB_PATHS[1]);
		end;
		if toolset.target == "avr" then
			dbg:set_param("GDB_SERVER", "run_avr -g ${GDB_PORT} ${TARGET_PATH}");
		elseif cmd.debug_type == "zephyr" then
			local debug_adapter_name = menu_bar.get_cmake_debug_adapter();
			local zephyr_debugserver = "\""..toolset.tools.west.."\"";
			zephyr_debugserver = zephyr_debugserver.. " debugserver --build-dir \""..cmd.build_path.."\" --skip-rebuild";
			-- pyocd, linkserver, jlink, nxp_s32dbg, arc-nsim, nios2
			if debug_adapter_name ~= "auto" then
				zephyr_debugserver = zephyr_debugserver.. " -r "..debug_adapter_name;
			end;
			zephyr_debugserver = zephyr_debugserver.. " --gdb-port ${GDB_PORT}";
			if debug_adapter_name == "openocd" then
				zephyr_debugserver = zephyr_debugserver.. " --telnet-port ${TELNET_PORT}";
				zephyr_debugserver = zephyr_debugserver.. " --tcl-port disabled";
			elseif debug_adapter_name == "pyocd" then
				zephyr_debugserver = zephyr_debugserver.. " --telnet-port ${TELNET_PORT}";
			end;
			print(zephyr_debugserver);
			dbg:set_param("OPENOCD", zephyr_debugserver);
			dbg:set_param("MAIN", "");
			dbg:set_param("GDB_INIT", make_path(cmd.build_path, ".gdbinit"));
			dbg:set_param("GDB_INIT_PRE", "set remotetimeout 60000\n");
			dbg._on_load_go = function(self)
				self:exec("load\n");
				self:go();
			end;
		elseif embeded_dbg_servers[cmd.debug_type] and toolset.debug_adapter then
			local load_svd;
			if toolset and toolset.find_svd then
				load_svd = function(device)
					local svd_path = toolset:find_svd(device);
					if svd_path == nil then
						return;
					end;
					print("Load SVD: ", svd_path);
					local peripherals = edx:get_object("pane:peripherals");
					peripherals:load_svd(svd_path);
				end;
			else
				find_svd = function()
					return nil;
				end;
			end;
			
			local is_segger_rtt_enabled = function()
				local cmake_arguments = project:get_scoped_setting(".cmake", "arguments");
				if not cmake_arguments then
					return false;
				end;
				local is_enable = {
					["1"] = true;
					["on"] = true;
					["yes"] = true;
					["true"] = true;
				};
				for idx, val in ipairs(cmake_arguments) do
					local enable_segger_rtt = val:match("^%-DENABLE_SEGGER_RTT=(.-)%s*$");
					if enable_segger_rtt and is_enable[enable_segger_rtt:lower()] then
						return true;
					end;
				end;
				return false;
			end;
			local find_target_ram_info = function(target)
				for group, ginfo in pairs(get_EMBEDED_CPU_DB()) do
					for idx, dev in ipairs(ginfo) do
						if dev[1] == target then
							for idx, mem_info in ipairs(dev[3]) do
								if mem_info[3] == 0 and mem_info[2] > 2048 then
									return mem_info[1], mem_info[2]
								end;
							end;
						end;
					end;
				end;
			end;
			local debug_adapter_name = menu_bar.get_cmake_debug_adapter();
			local debug_adapter = toolset.debug_adapter;
			local debug_adapter_info = debug_adapter.adapters[debug_adapter_name];
			local debug_adapter_speed = (#debug_adapter.speeds>0) and menu_bar.get_cmake_debug_adapter_speed();
			local debug_adapter_transport = (#debug_adapter.transports>0) and menu_bar.get_cmake_debug_adapter_transport();
			if debug_adapter_info["type"] == "openocd" then
				local openocd_scripts = debug_adapter_info.scripts;
				local openocd_path = debug_adapter_info.bin or nil;
				if openocd_path == nil or utils:file_time(openocd_path) == nil then
					show_notify(LANG("notify/debug/openocd_not_found"));
					return false;
				end;
				local openocd_target, device_name = project:find_debug_target(toolset, debug_adapter_name);
				local openocd = "\""..openocd_path.."\"";
				openocd = openocd.." --search \""..make_path(openocd_scripts).. "\"";
				openocd = openocd.." -f interface/"..debug_adapter_info.interface.. ".cfg";

				debug_adapter_transport = debug_adapter_transport and debug_adapter_transport:lower();
				if debug_adapter_info.interface == "stlink" then
					if debug_adapter_transport == "swd" then
						debug_adapter_transport = "hla_swd";
					elseif debug_adapter_transport == "jtag" then
						debug_adapter_transport = "hla_jtag";
					end;
				end;
				
				if debug_adapter_transport then
					openocd = openocd.." -c \"transport select ".. debug_adapter_transport .. "\"";
				end;
				if debug_adapter_speed then
					openocd = openocd.." -c \"adapter speed "..debug_adapter_speed:lower().."\"";
				end;
				openocd = openocd.." -f target/"..openocd_target.. ".cfg";
				openocd = openocd.." -c \"gdb_port ${GDB_PORT}\"";
				openocd = openocd.." -c \"telnet_port ${TELNET_PORT}\"";
				openocd = openocd.." -c \"tcl_port disabled\"";
				print(openocd);
				dbg:set_param("OPENOCD", openocd);
				dbg:set_param("MAIN", "");
				dbg:set_param("GDB_INIT", make_path(cmd.build_path, ".gdbinit"));
				dbg:set_param("GDB_INIT_PRE", "set remotetimeout 60000\n");
				dbg._on_load_go = function(self)
					if toolset["type"] == "esp-idf" then
						self:exec("monitor reset halt\n");
					else
						-- if toolset.target:match("^arm.*") then
						-- 	self:exec("monitor arm semihosting enable\n");
						-- end;
						self:exec("load\n");
						local target_ram_begin, target_ram_size = find_target_ram_info(device_name);
						-- RTT support
						if target_ram_begin and target_ram_size then
							if target_ram_size > 2048 then
								target_ram_size = 2048;	-- searching for segger rtt in top 2K ram
							end;
							local set_rtt_range = ("mon rtt setup 0x%X 0x%X \"SEGGER RTT\""):format(target_ram_begin, target_ram_size);
							self:exec(set_rtt_range);
							if is_segger_rtt_enabled() then
								self._on_locate_file = function(self)
									self:exec("mon rtt start");
								end;
							end;
						end;
					end;
					load_svd(device_name);
					self:go();
				end;
			elseif debug_adapter_info["type"] == "jlink" then
				-- https://wiki.segger.com/J-Link_GDB_Server#Command_line_options
				local jlinkgdb_path = debug_adapter_info.bin;
				if jlinkgdb_path == nil or utils:file_time(jlinkgdb_path) == nil then
					show_notify(LANG("notify/debug/jlinkgdb_not_found"));
					return false;
				end;
				local debug_target_device, device_name = project:find_debug_target(toolset, debug_adapter_name);
				local jlinkdbg = "\""..jlinkgdb_path.."\" -halt";
				jlinkdbg = jlinkdbg.." -if "..debug_adapter_transport;
				jlinkdbg = jlinkdbg.." -device "..debug_target_device;
				jlinkdbg = jlinkdbg.." -speed "..debug_adapter_speed;
				jlinkdbg = jlinkdbg.." -port ${GDB_PORT}";
				jlinkdbg = jlinkdbg.." -rtttelnetport "..segger_rtt_port;
				jlinkdbg = jlinkdbg.." -telnetport ${TELNET_PORT}";
				print(jlinkdbg);
				dbg:set_param("OPENOCD", jlinkdbg);
				dbg:set_param("MAIN", "");
				dbg:set_param("GDB_INIT", make_path(cmd.build_path, ".gdbinit"));
				dbg:set_param("GDB_INIT_PRE", "set remotetimeout 60000\n");
				dbg._on_load_go = function(self)
					if toolset["type"] == "esp-idf" then
						self:exec("monitor reset halt\n");
					else
						-- if toolset.target:match("^arm.*") then
						-- 	self:exec("monitor semihosting enable\n");
						-- end;

						local target_ram_begin, target_ram_size = find_target_ram_info(device_name);
						if target_ram_begin and target_ram_size then
							local set_rtt_range = ("mon exec SetRTTSearchRanges 0x%X 0x%X"):format(target_ram_begin, target_ram_size);
							self:exec(set_rtt_range);
							self:set_param("ENABLE_SEGGER_RTT_MONITOR", tostring(segger_rtt_port));
						end;
						self:exec("monitor reset\n");
						self:exec("load\n");
					end;
					load_svd(device_name);
					self:go();
				end;
			elseif debug_adapter_info["type"] == "stlink" then
				local stlinkgdb_path = debug_adapter_info.bin;
				if stlinkgdb_path == nil or utils:file_time(stlinkgdb_path) == nil then
					show_notify(LANG("notify/debug/stinkgdb_not_found"));
					return false;
				end;
				local debug_target_device = toolset.openocd_target or toolset.target;
				local stlinkdbg = "\""..stlinkgdb_path.."\" --halt";
				if debug_adapter_transport == "SWD" then
					stlinkdbg = stlinkdbg.." --swd";
				end;
				-- stlinkdbg = stlinkdbg.." -device "..debug_target_device;
				stlinkdbg = stlinkdbg.." --frequency "..debug_adapter_speed;
				stlinkdbg = stlinkdbg.." --stm32cubeprogrammer-path \""..make_path(debug_adapter_info.prog_bin, "..").."\"";
				stlinkdbg = stlinkdbg.." --pend-halt-timeout 30";
				stlinkdbg = stlinkdbg.." --port-number ${GDB_PORT}";
				-- stlinkdbg = stlinkdbg.." --attach"; -- attaches to the running target
				-- stlinkdbg = stlinkdbg.." --swo-port ${TELNET_PORT}"; -- for client print
				print(stlinkdbg);
				dbg:set_param("OPENOCD", stlinkdbg);
				dbg:set_param("MAIN", "");
				dbg:set_param("GDB_INIT", make_path(cmd.build_path, ".gdbinit"));
				dbg:set_param("GDB_INIT_PRE", "set remotetimeout 60000\n");
				dbg._on_load_go = function(self)
					if toolset["type"] == "esp-idf" then
						self:exec("monitor reset halt\n");
					else
						self:exec("load\n");
					end;
					load_svd(device_name);
					self:go();
				end;
			elseif debug_adapter_info["type"] == "pyocd" then
				local pyocd_path = debug_adapter_info.bin;
				local pyocd = "\""..pyocd_path.."\" gdb";
				local pyocd_target, device_name = project:find_debug_target(toolset, debug_adapter_name);
				pyocd = pyocd.." -t "..pyocd_target;
				pyocd = pyocd.." -f "..debug_adapter_speed.."k";
				if debug_adapter_transport then
					pyocd = pyocd.." -O dap_protocol=".. debug_adapter_transport;
				end;
				pyocd = pyocd.." --port ${GDB_PORT}";
				pyocd = pyocd.." --telnet-port ${TELNET_PORT}";
				print(pyocd);
				dbg:set_param("OPENOCD", pyocd);
				dbg:set_param("MAIN", "");
				dbg:set_param("GDB_INIT", make_path(cmd.build_path, ".gdbinit"));
				dbg:set_param("GDB_INIT_PRE", "set remotetimeout 60000\n");
				dbg._on_load_go = function(self)
					if toolset["type"] == "esp-idf" then
						self:exec("monitor reset halt\n");
					else
						-- if toolset.target:match("^arm.*") then
						-- 	self:exec("monitor arm semihosting enable\n");
						-- end;
						local target_ram_begin, target_ram_size = find_target_ram_info(device_name);
						-- RTT support
						if target_ram_begin and target_ram_size then
							if target_ram_size > 2048 then
								target_ram_size = 2048;	-- searching for segger rtt in top 2K ram
							end;
							local set_rtt_range = ("mon rtt setup 0x%X 0x%X \"SEGGER RTT\""):format(target_ram_begin, target_ram_size);
							self:exec(set_rtt_range);
							if is_segger_rtt_enabled() then
								self._on_locate_file = function(self)
									self:exec("mon rtt start");
								end;
							end;
						end;
						self:exec("load\n");
					end;
					load_svd(device_name);
					self:go();
				end;
			elseif debug_adapter_info["type"] == "t-head" then
				local thead_path = debug_adapter_info.bin;
				local thead = "\""..thead_path.."\"";
				-- local thead_target, device_name = project:find_debug_target(toolset, debug_adapter_name);
				if debug_adapter_speed then
					thead = thead.." -setclk "..debug_adapter_speed.."k";
				end;
				if debug_adapter_transport then
				end;
				-- thead = thead.." -local-semi";
				thead = thead.." -port ${GDB_PORT}";
				print(thead);
				dbg:set_param("OPENOCD", thead);
				dbg:set_param("MAIN", "");
				dbg:set_param("GDB_INIT", make_path(cmd.build_path, ".gdbinit"));
				dbg:set_param("GDB_INIT_PRE", "set remotetimeout 60000\n");
				dbg._on_load_go = function(self)
					self:exec("monitor reset\n");
					load_svd(device_name);
					self:go();
				end;
			end;
		end;
	end;
	if toolset then
		if toolset.wsl then
			dbg_target = string.format("wsl://[%s]%s", toolset.wsl, dbg_target);
		elseif toolset.remote then
			dbg_target = string.format("%s%s", toolset.remote, dbg_target);
		end;
	end;
	if not dbg_target then
		print("debugee not specified!");
		return;
	end;

	dbg.__dbg_target = dbg_target;
	dbg.__toolset = toolset;
	if not dbg:load(dbg_target) then
		if utils:file_time(dbg_target) then
			show_notify(LANG("notify/debug/invalid_target") .. "\n" .. dbg_target);
		else
			show_notify(LANG("notify/debug/target_not_exists") .. "\n" .. dbg_target);
		end;
	end;
end;

local function add_debug_map_path(path_map)
	local dbg = edx.dbg;
	if dbg == nil then
		print("no debugger!");
		return false;
	end;
	if type(path_map) ~= "table" then
		print("invalid parameter!");
		return false;
	end;
	local source_map = dbg.__source_map;
	if source_map == nil then
		local function do_map_path(p, from_path, to_path, from_split, to_split)
			local p_prefix = string.sub(p, 1, #from_path);
			if p_prefix == from_path then
				local ext_path = string.sub(p, #from_path + 1);
				return to_path..ext_path:gsub(from_split, to_split);
			end;
			return nil;
		end;
		source_map = {};
		local old_map_path = dbg.map_path or function(p) return p; end;
		local old_unmap_path = dbg.unmap_path or function(p) return p; end;
		dbg.__source_map = source_map;
		dbg.map_path = function(p)
			for idx, map in ipairs(source_map) do
				local local_path = map[1];
				local remote_path = map[2];
				remote_path = do_map_path(p, local_path, remote_path, "\\", "/");
				if remote_path then
					return remote_path;
				end;
			end;
			return old_map_path(p);
		end;
		dbg.unmap_path = function(p)
			for idx, map in ipairs(source_map) do
				local local_path = map[1];
				local remote_path = map[2];
				local_path = do_map_path(p, remote_path, local_path, "/", "\\");
				if local_path then
					return edx:regular_path(local_path);
				end;
			end;
			return old_unmap_path(p);
		end;
	end;
	if type(path_map[1]) == "string" then
		local remote_path = path_map[2];
		if type(remote_path) ~= "string" then
			return false, "invalid remote path";
		end;
		path_map[1] = edx:regular_path(path_map[1]) or path_map[1];
		table.insert(source_map, path_map);
	else
		for idx, map in ipairs(path_map) do
			map[1] = edx:regular_path(map[1]) or map[1];
			table.insert(source_map, map);
		end;
	end;
	if dbg.is_running() then
		dbg:locate_file();
	end;
end;

function debug_program(target, dbg_type, dbg_argument, dbg_working_path, dbg_environment)
	if dbg_type == nil then
		dbg_type = "windbg";
	end;
	edx:select_debugger(dbg_type);
	local dbg = get_dbg();
	dbg._post_load_handler = {};
	if dbg.is_running() then
		print("debugger is running");
		return;
	end;
	if dbg_environment then
		for key, value in pairs(dbg_environment) do
			dbg:set_environment(key, value);
		end;
	end;
	if dbg_working_path then
		if dbg_working_path:sub(-1) == "\\" and #dbg_working_path > 3 then
			dbg_working_path = dbg_working_path:sub(1,-2);
		end;
		dbg:set_working_path(dbg_working_path);
	end;
	if dbg_argument then
		dbg:set_args(dbg_argument);
	end;
	if dbg_type:match("gdb/") then
		local gdb_post_load_handler = function()
			if target then
				print("debug target:", target, " ", dbg_argument);
				if dbg_working_path then
					print("debug cwd:", dbg_working_path);
				end;
			end;
			if toolset then
				if toolset.wsl then
					dbg:set_param("ENABLE_WSL_CONSOLE", toolset.wsl);
				elseif toolset.remote then
					dbg:set_param("ENABLE_SSH_CONSOLE", toolset.remote);
				end;
			end;
		end;
		table.insert(dbg._post_load_handler, gdb_post_load_handler);
		if GDB_PATHS and GDB_PATHS[1] then
			dbg:set_param("GDB_PATH", GDB_PATHS[1]);
		end;
	end;

	local ssh_target_info = {string.match(target, [[ssh://.-(/.+/%.edx)/(%w)/(.+)/%.edx/cmake/.+]])};
	if ssh_target_info[1] then
		local local_path = ssh_target_info[2]..[[:\]]..ssh_target_info[3];
		local_path = edx:regular_path(local_path) or local_path;
		local_path = local_path:gsub("/", "\\");
		local remote_path = ssh_target_info[1]..[[/]]..ssh_target_info[2]..[[/]]..ssh_target_info[3];
		print("map source path:", remote_path, " => ", local_path);
		add_debug_map_path({local_path, remote_path});
	else
		dbg.map_path = function(p) return p; end;
		dbg.unmap_path = function(p) return p; end;
	end;
	dbg.dbg_type = dbg_type;
	dbg.beautifier = edx.get_debugger_beautifier(dbg_type);
	if not dbg:load(target) then
		if utils:file_time(target) then
			print(LANG("notify/debug/invalid_target") .. " " .. target);
		else
			print(LANG("notify/debug/target_not_exists") .. " " .. target);
		end;
	end;
end;

function debug_map_path(path_map)
	local dbg = edx.dbg;
	if dbg == nil or dbg.is_running() == false then
		print("no active debugger!");
		return false;
	end;
	add_debug_map_path(path_map);
end;

function debug_stop()
	local dbg_obj = edx.dbg;
	if dbg_obj ~= nil and dbg_obj:is_running() then
		dbg_obj:terminate();
		return true;
	end;
	return false;
end;

function is_debugger_running()
	return edx.dbg and edx.dbg:is_running();
end;

global_cmd_map[_op.debug_start] = function(self,id)
	local dbg_obj = get_dbg();
	if dbg_obj:is_running() then
		if dbg_obj.last_fp ~= nil then
			dbg_obj.last_fp[1].set_dbg_cursor(-1);
		end;
		menu_bar:switch_debug_menu("debugging");
		dbg_obj:go();
	else
		print("run");
		load_program({});
	end;
	return 1;
end;

global_cmd_map[_op.debug_terminate] = function(self,id)
	if debug_stop() then
		print("terminate");
	end;
	return 1;
end;

local function _set_breakpoint_on_main(dbg)
	local breakpoint_symbols = {};
	local toolset = dbg.__toolset;
	if utils:file_time(dbg.__dbg_target) then
		local is_gui_exe = false;
		local is_msvc = false;
		if toolset and toolset["type"] == "msvc" then
			is_msvc = true;
		end;

		local type_id, type_desc = edx:guess_file_type(dbg.__dbg_target);
		if type_desc then
			if is_msvc and type_desc:match("^PE.*GUI.*$") then
				-- msvc target set breakpoint on WinMain at first
				breakpoint_symbols = {"wWinMain", "WinMain", "main"};
			elseif type_desc:match("^PE.*$") then
				breakpoint_symbols = {"main", "WinMain", "wWinMain"};
			end;
		end;
	end;

	if #breakpoint_symbols == 0 and toolset then
		if toolset["type"] == "esp-idf" then
			breakpoint_symbols = {"app_main"};
		else
			breakpoint_symbols = {"main"};
		end;
	end;

	for idx, symbol_name in ipairs(breakpoint_symbols) do
		if dbg.set_breakpoint_ex({flags = 0x11;name = symbol_name;}) ~= 0 then
			return;
		end;
	end;
	return 0;
end;

global_cmd_map[_op.debug_step_over] = function(self,id)
	-- print( "step" );
	local dbg_obj = get_dbg();
	if dbg_obj:is_running() then
		menu_bar:switch_debug_menu("debugging");
		dbg_obj:step();
	else
		load_program({_set_breakpoint_on_main});
	end;
	return 1;
end;

global_cmd_map[_op.debug_step_in] = function(self,id)
	-- print( "trace into" );
	local dbg_obj = get_dbg();
	if dbg_obj:is_running() then
		menu_bar:switch_debug_menu("debugging");
		dbg_obj:trace();
	else
		load_program({_set_breakpoint_on_main});
	end;
	return 1;
end;

global_cmd_map[_op.debug_step_out] = function(self,id)
	-- print( "trace return" );
	local dbg_obj = get_dbg();
	if dbg_obj:is_running() then
		menu_bar:switch_debug_menu("debugging");
		dbg_obj:trace_return();
	else
		load_program();
	end;
	return 1;
end;

function dbg_toggle_breakpoint(doc, line)
	-- print( "toggle breakpoint" );
	local dbg_obj = edx.dbg;
	local name = bkpt_file_id(doc.file_name);

	if dbg_obj == nil then
		doc:toggle_breakpoint(line);
		return;
	end;

	local bkps = dbg_obj.bkps;
	if bkps == nil then
		bkps = {};
		dbg_obj.bkps = bkps;
	end;
	if bkps[name] == nil then
		bkps[name] = {};
	end;

	if doc:toggle_breakpoint(line) then
		local bkp_id = dbg_obj:set_breakpoint_mapped(name, line + 1);
		if bkp_id == 0 then
			print("Can't set breakpoint!");
			doc:toggle_breakpoint(line);
		else
			bkps[name][tostring(line+1)] = bkp_id;
		end;
	elseif bkps[name][tostring(line+1)] ~= nil then
		if dbg_obj:del_breakpoint(bkps[name][tostring(line+1)]) then
			bkps[name][tostring(line+1)] = nil;
		else
			print("Can't delete breakpoint!");
			doc:toggle_breakpoint(line);
		end;
	end;
	return 1;
end;

global_cmd_map[_op.debug_toggle_breakpoint] = function(self, id)
	local doc = mgr.current_document;
	dbg_toggle_breakpoint(doc, doc.cursor_line);
end;

global_cmd_map[_op.debug_break] = function(self,id)
	-- print("pause");
	local dbg_obj = edx.dbg;
	if dbg_obj ~= nil and dbg_obj:is_running() then
		dbg_obj:pause();
	end;
	return 1;
end;

global_cmd_map[_op.debug_jump_to] = function(self,id)
	local dbg_obj = edx.dbg;
	local doc = mgr.current_document;
	local name = doc.file_name;
	local line = doc.cursor_line + 1;
	print("jump ", name, ":", line);
	if dbg_obj ~= nil and dbg_obj:is_running() then
		dbg_obj:jump(dbg_obj.map_path(name), line);
	end;
	return 1;
end;

global_cmd_map[_op.debug_run_to_cursor] = function(self,id)
	local dbg_obj = edx.dbg;
	local doc = mgr.current_document;
	local name = doc.file_name;
	local line = doc.cursor_line + 1;
	print("runto cursor ", name, ":", line);
	if dbg_obj ~= nil and dbg_obj:is_running() then
		if dbg_obj:set_breakpoint_ex{flags = 0x10; ["file"] = dbg_obj.map_path(name); ["line"] = line;} ~= 0 then
			dbg_obj:go();
		end;
	end;
	return 1;
end;

menu_bar:switch_debug_menu("stopped");
print("load debugger binding ok!");

-- 加载 debugger beautifiers
dbg_call(function() require "windbg_beautifier" end, "loading windbg beautifier failed!");
